-----------------------------------------------------------
-- locals
local lg    = love.graphics
local floor = math.floor
local ceil  = math.ceil
local min   = math.min
local max   = math.max
-- use fixed font for drawing cells
local font        = lg.newFont(12)
local fmidheight  = font:getHeight()/2
local whiteColor  = {255,255,255}
-----------------------------------------------------------
-- map class
local m   = {class = 'map'}
m.__index = m
m.new     = function(self,t)
	return setmetatable(t,self)
end
m.__call  = m.new
setmetatable(m,{__call = m.new})
-----------------------------------------------------------
-- draw map
function m.draw(map)
	local prevFont  = lg.getFont()
	local w,h       = map.tilewidth,map.tileheight
	local wc,hc     = w/2,h/2
	-- draw each cell with tile label and color
	lg.setFont(font)
	for y,row in pairs(map.grid) do
    for x,id in pairs(row) do
      lg.setColor(map.tileset[id].color or whiteColor)
      -- first cell in positive quadrant starts at (1,1)
      local x,y = (x-1)*w,(y-1)*h
      lg.rectangle('line',x,y,w,h)
      x,y = x+wc-font:getWidth(id)/2,y+hc-fmidheight
      lg.print(id,x,y)
    end
	end
	lg.setFont(prevFont or font)	
end
-----------------------------------------------------------
-- setters
function m:setTileID(x,y,id)
	self.grid[y][x] = id
end

function m:setTileSize(w,h)
	self.tileheight = h
	self.tilewidth  = w
end

function m:setGrid(grid)
	self.grid = grid
end

function m:setTileset(tileset)
	self.tileset = tileset
end

function m:setCallbacks(preSolve,postSolve)
	self.preSolve = preSolve; self.postSolve = postSolve
end
-----------------------------------------------------------
-- getters
function m:getTile(x,y)
	return self.tileset[ self.grid[y][x] ]
end

function m:getTileSize()
	return self.tilewidth,self.tileheight
end

function m:getRange(x,y,w,h)
	local mw,mh = self:getTileSize()
	local gx,gy =  
		floor(x/mw)+1,
		floor(y/mh)+1
	local gx2,gy2 =
		ceil((x+w)/mw),
		ceil((y+h)/mh)
	return gx,gy,gx2,gy2
end
-----------------------------------------------------------
-- collision callback when tile overlaps with sensor
-- return true if collision should be resolved
function m:preSolve(side,obj,gx,gy)
end

-----------------------------------------------------------
-- callback when collision is resolved
function m:postSolve(side,obj,gx,gy)
end
-----------------------------------------------------------
-- resolve horizontal collisions
function m:resolveXdirection(obj,x,y,w,h)
	local mw,mh         = self:getTileSize()
	local gx,gy,gx2,gy2 = self:getRange(x,y,w,h)

	-- right sensor check
	for gy = gy,gy2 do
		-- do collision if callback returns true
		if self:preSolve('right',obj,gx2,gy) then
			x = ((gx2-1)*mw)-w
			self:postSolve('right',obj,gx2,gy)
		end
	end

	-- left sensor check
	for gy = gy,gy2 do
		if self:preSolve('left',obj,gx,gy) then
			x = gx*mw
			self:postSolve('left',obj,gx,gy)
		end
	end
	return x
end

-----------------------------------------------------------
-- resolve vertical collisions
function m:resolveYdirection(obj,x,y,w,h)
	local mw,mh         = self:getTileSize()
	local gx,gy,gx2,gy2 = self:getRange(x,y,w,h)

	-- floor sensor check
	for gx = gx,gx2 do
		if self:preSolve('floor',obj,gx,gy2) then
			y = (gy2-1)*mh-h
			self:postSolve('floor',obj,gx,gy2)
		end
	end

	-- ceiling sensor check
	for gx = gx,gx2 do
		if self:preSolve('ceiling',obj,gx,gy) then
			y = gy*mh
			self:postSolve('ceiling',obj,gx,gy)
		end
	end
	return y
end
-----------------------------------------------------------
return m